/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * Originally written by Alexander Vdolainen <avdolainen@gmail.com>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <getopt.h>
#include <syslog.h>
#include <pthread.h>
#include <sexpr/sexp.h>

#include <dsbus/data.h>
#include <dsbus/daemon.h>


static void __print_version(void)
{
  fprintf(stdout, "DSB user space daemon version: 0.1\n\n");
  fprintf(stdout, "Copyright (c) 2015-2016 Alexander Vdolainen.\nEmail: avdolainen@gmail.com\n\n");

  return;
}

static void __print_help(void)
{
  fprintf(stdout, "Usage: dsbusd [options]\nOptions:\n");
  fprintf(stdout, " --daemonize(-d)\t Daemonize at startup for normal init system.\n");
	fprintf(stdout, " --systemd(-s)\t Daemonize adopted for idiots init system i.e. systemd.\n");
  fprintf(stdout, " --config-file-path(-c) <file>\t Use <file> as the configuration file.\n");
  fprintf(stdout, " --unix-socket-path(-u) <file>\t Use <file> as the unix socket name.\n");
	fprintf(stdout, " --log-file-path(-l) <file>\t Use <file> as the logging output.\n");
  fprintf(stdout, " --help(-h)\t Print there help message\n");
  fprintf(stdout, " --version(-v)\t Print version information\n");
  fprintf(stdout, "\nFor further information check out provided documentation.\n");
  return;
}

/* this version of skeleton daemon targeted for the normal init system
 * i.e. sysvinit, openrc, bsd-style init etc...
 * it will works with systemd like other usual *nix daemon.
 * there are no sense to make a lot of code for each weird init system
 */
static void skeleton_daemon()
{
  pid_t pid;
  int x;

  /* Fork off the parent process */
  pid = fork();

  /* An error occurred */
  if(pid < 0)
      exit(EXIT_FAILURE);

  /* Success: Let the parent terminate */
  if(pid > 0)
    exit(EXIT_SUCCESS);

  /* On success: The child process becomes session leader */
  if(setsid() < 0)
    exit(EXIT_FAILURE);

  /* Catch, ignore and handle signals */
  signal(SIGCHLD, SIG_IGN);
  signal(SIGHUP, SIG_IGN);

  /* Fork off for the second time*/
  pid = fork();

  /* An error occurred */
  if(pid < 0)
     exit(EXIT_FAILURE);

  /* Success: Let the parent terminate */
  if(pid > 0)
    exit(EXIT_SUCCESS);

  /* TODO: FIXME: Set new file permissions */
  //umask(0);

  /* TODO: FIXME: Change the working directory to the root directory */
  /* or another appropriated directory */
  //chdir("/");

  /* Close all open file descriptors */
  for(x = sysconf(_SC_OPEN_MAX); x>0; x--) {
    close(x);
  }

  /* Open the log file */
  openlog("dsbusd", LOG_PID, LOG_DAEMON);

  return;
}

static void *__unixsocket_thread(void *arg)
{
	struct dsbd_ctx *gctx = (struct dsbd_ctx *)arg;
	int s, len, s2;
	socklen_t t;
  struct sockaddr_un local, remote;
	struct _payload *pload;
	char buf[4096];

	if(!gctx) return NULL;

	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  local.sun_family = AF_UNIX;
  strcpy(local.sun_path, (const char *)gctx->unixsck_path);
  unlink(local.sun_path);
  len = strlen(local.sun_path) + sizeof(local.sun_family);
  if (bind(s, (struct sockaddr *)&local, len) == -1) {
		syslog(LOG_CRIT, "Unable to create a unix socket for cmds.\n");
		return NULL;
  }

  if (listen(s, MAX_BACKLOG_CONNECTIONS) == -1) {
		syslog(LOG_CRIT, "Unable to listen a unix socket for cmds.\n");
		return NULL;
  }

	for(;;) {
    int done, n;
    t = sizeof(remote);
    if ((s2 = accept(s, (struct sockaddr *)&remote, &t)) == -1) {
			syslog(LOG_CRIT, "Unable to accept a unix socket connection for cmds.\n");
			return NULL;
    }

    done = 0;
    do {
      n = recv(s2, buf, sizeof(char)*4096, 0);
      if (n <= 0) {
        if (n < 0) perror("recv");
        done = 1;
      }

      if (!done) {
				/* let's deal with that */
				pload = malloc(sizeof(struct _payload));
				if(!pload) {
					syslog(LOG_CRIT, "Cannot allocate memory for message from unix socket.\n");
				} else {
					pload->pmsg = strdup(buf);
					if(pload->pmsg) {
						pload->scn = s2;
						if((n = zst_queue_add(gctx->uqueue, pload, USR_MSG))) {
							free(pload->pmsg);
							free(pload);
							close(s2);
						}
					}
				}
				done = 1;
      }
    } while (!done);

    memset(buf, 0, sizeof(char)*4096);
  }

	return NULL;
}

static void __print_active_object_from_storage(FILE *out,
																							 dsb_secstore_t *storage)
{
	usrtc_t *stree = &(storage->curr_data);
  usrtc_node_t *node = NULL;
  dsb_secinfo_t *info = NULL;
	char *uus;

  pthread_rwlock_rdlock(&(storage->srwlock));

	if(!(node = usrtc_first(stree))) {
		fprintf(out, "(No data).\n");
		goto __unlock;
	}

  for(node = usrtc_first(stree); ;
      node = usrtc_next(stree, node)) {
    info = (dsb_secinfo_t *)usrtc_node_getdata(node);

    pthread_mutex_lock(&info->slock);

    if(!(info->flags & DSB_GARBAGED)) {
			uus = uuid2str(info->blk_uuid);
			fprintf(out, "%lu: @%s inum: %lu name: '%s'\n", info->hash,
							uus ? uus : "nil", info->inode_num,
							info->obj_name);
			if(uus) free_uuidstr(uus);
		}

    pthread_mutex_unlock(&info->slock);
		if(node == usrtc_last(stree)) break;
  }

 __unlock:
  pthread_rwlock_unlock(&(storage->srwlock));

	return;
}

static void __print_object_entries(FILE *out, dsb_secstore_t *storage,
																	 uint64_t hash)
{
	dsb_secinfo_t *info;
	int r = 0;
	char *uus;

	r = dsb_store_search(storage, hash, &info);
	if(r) fprintf(out, "Object not found, or internal error occurs (%d).\n", r);
	else {
		pthread_mutex_lock(&info->slock);
		uus = uuid2str(info->blk_uuid);
		fprintf(out, "%lu: @%s inum: %lu name: '%s'\n", info->hash,
						uus ? uus : "nil", info->inode_num,
						info->obj_name);
		if(uus) free_uuidstr(uus);

		fprintf(out, "UID List (active):\n");

		for(r = 0; r < __MXRLE; r++) {
			if(info->aclist.uidlist[r].mask & DSBMASK_SET) {
				fprintf(out, "\t n: %d UID: %d Mode: ", r, info->aclist.uidlist[r].uid);
				/* human output */
				if(info->aclist.uidlist[r].mask & DSBMASK_READ) fprintf(out, "r");
				else fprintf(out, "-");
				if(info->aclist.uidlist[r].mask & DSBMASK_WRITE) fprintf(out, "w");
				else fprintf(out, "-");
				if(info->aclist.uidlist[r].mask & DSBMASK_EXEC) fprintf(out, "x");
				else fprintf(out, "-");
				fprintf(out, "\n");
			}
		}

		fprintf(out, "\n");

		fprintf(out, "GID List (active):\n");
		for(r = 0; r < __MXRLE; r++) {
			if(info->aclist.gidlist[r].mask & DSBMASK_SET) {
				fprintf(out, "\t n: %d GID: %d Mode: ", r, info->aclist.gidlist[r].gid);
				/* human output */
				if(info->aclist.gidlist[r].mask & DSBMASK_READ) fprintf(out, "r");
				else fprintf(out, "-");
				if(info->aclist.gidlist[r].mask & DSBMASK_WRITE) fprintf(out, "w");
				else fprintf(out, "-");
				if(info->aclist.gidlist[r].mask & DSBMASK_EXEC) fprintf(out, "x");
				else fprintf(out, "-");
				fprintf(out, "\n");
			}
		}

		fprintf(out, "\n");

		fprintf(out, "Mode for others: ");
		if(info->aclist.othmask & DSBMASK_SET) {
			/* human output */
			if(info->aclist.othmask & DSBMASK_READ) fprintf(out, "r");
			else fprintf(out, "-");
			if(info->aclist.othmask & DSBMASK_WRITE) fprintf(out, "w");
			else fprintf(out, "-");
			if(info->aclist.othmask & DSBMASK_EXEC) fprintf(out, "x");
			else fprintf(out, "-");
			fprintf(out, "\n\n");
		} else fprintf(out, "Not set.\n\n");

		pthread_mutex_unlock(&info->slock);
	}

	return;
}

static void __add_object_storage(FILE *out, dsb_secstore_t *storage, char *name,
																 uint64_t inum, uint8_t blkid[16])
{
	dsb_secinfo_t *info = malloc(sizeof(dsb_secinfo_t));
	char str[512], *uc = NULL;
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	if(!info) fprintf(out, "Not enough memory.\n");

	r = dsb_store_info_allocinit(info);
	if(r) fprintf(out, "Something goes wrong with %d\n", r);

	/* init */
	strncpy(info->obj_name, name, 256);
	memcpy(info->blk_uuid, blkid, sizeof(uint8_t)*16);
	info->inode_num = inum;

	/* ok, take a deal with a hash:
	 * it's forms from the following:
	 * uuid-inode_num-name
	 */
	uc = uuid2str(blkid);
	snprintf(str, 512, "%s-%lu", uc ? uc : "nil", inum);
	free_uuidstr(uc);

	info->hash = (uint64_t)djb2_hash((unsigned char *)str);

	r = dsb_store_insert(storage, info);

	if(r) fprintf(out, "Something goes wrong with %d\n", r);
	else {
		/* form a message for kernel */
		memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
		dsbmsg.cmd = DSBNL_CMDADDOBJ;
		dsbmsg.hash = info->hash;
		r = netlink_senddsbmsg(storage, &dsbmsg);
		fprintf(out, "(%d)DONE.\n", r);
	}

	return;
}

static void __remove_object_storage(FILE *out, dsb_secstore_t *storage,
																		uint64_t hash)
{
	int r = dsb_store_remove(storage, hash);
	struct dsbnl_policy_msg dsbmsg;

	if(r) fprintf(out, "Something goes wrong with %d\n", r);
	else {
		/* form a message for kernel */
		memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
		dsbmsg.cmd = DSBNL_CMDREMOVEOBJ;
		dsbmsg.hash = hash;
		r = netlink_senddsbmsg(storage, &dsbmsg);
		fprintf(out, "(%d)DONE.\n", r);
	}

	return;
}

static void __setrule_object(FILE *out, dsb_secstore_t *storage, uint64_t hash,
														 int n, uint32_t id, int rid, uint8_t mask)
{
	dsb_secinfo_t *info;
	dsb_acl_entry_t *list;
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	if(n >= __MXRLE || n < 0) {
	__einval:
		fprintf(out, "INVALID(EINVAL)\n");
		return;
	}

	r = dsb_store_search(storage, hash, &info);
	if(r) fprintf(out, "Object not found, or internal error occurs (%d).\n", r);
	else {
		pthread_mutex_lock(&info->slock);
		switch(rid) {
		case DSBRID_UID: list = info->aclist.uidlist; break;
		case DSBRID_GID: list = info->aclist.gidlist; break;
		case DSBRID_OTHER:
			info->aclist.othmask = mask;
			info->aclist.othmask |= DSBMASK_SET;
			pthread_mutex_unlock(&info->slock);
			goto __sendmsg;
			break;
		}
		if(!list) goto __einval;
		list[n].uid = id;
		list[n].mask = mask;
		list[n].mask |= DSBMASK_SET;
		pthread_mutex_unlock(&info->slock);

	__sendmsg:
		/* now we need to call the dsb within kernel */
		memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
		dsbmsg.cmd = DSBNL_CMDSETLISTENTRY;
		dsbmsg.hash = info->hash;
		dsbmsg.listno = n;
		switch(rid) {
		case DSBRID_UID:
			dsbmsg.uid = id;
			break;
		case DSBRID_GID:
			dsbmsg.gid = id;
			break;
		}
		dsbmsg.flags = rid;
		dsbmsg.cap = mask;

		r = netlink_senddsbmsg(storage, &dsbmsg);
		fprintf(out, "(%d)DONE.\n", r);
	}

	return;
}

static void __cleanrule_object(FILE *out, dsb_secstore_t *storage, uint64_t hash,
															 int n, int rid)
{
	dsb_secinfo_t *info;
	dsb_acl_entry_t *list;
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	if(n >= __MXRLE || n < 0) {
	__einval:
		fprintf(out, "INVALID(EINVAL)\n");
		return;
	}

	r = dsb_store_search(storage, hash, &info);
	if(r) fprintf(out, "Object not found, or internal error occurs (%d).\n", r);
	else {
		pthread_mutex_lock(&info->slock);
		switch(rid) {
		case DSBRID_UID: list = info->aclist.uidlist; break;
		case DSBRID_GID: list = info->aclist.gidlist; break;
		case DSBRID_OTHER:
			info->aclist.othmask = 0;
			pthread_mutex_unlock(&info->slock);
			goto __sendmsg;
			break;
		}
		if(!list) goto __einval;
		list[n].uid = 0;
		list[n].mask = 0;
		pthread_mutex_unlock(&info->slock);

	__sendmsg:
		/* now we need to call the dsb within kernel */
		memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
		dsbmsg.cmd = DSBNL_CMDCLEANLISTENTRY;
		dsbmsg.hash = info->hash;
		dsbmsg.listno = n;
		dsbmsg.flags = rid;

		r = netlink_senddsbmsg(storage, &dsbmsg);
		fprintf(out, "(%d)DONE.\n", r);
	}

	return;
}

static void __usercap_set(FILE *out, dsb_secstore_t *storage, uid_t uid, uint8_t n)
{
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
	dsbmsg.cmd = DSBNL_CMDUIDCAPADD;
	dsbmsg.uid = uid;
	dsbmsg.listno = n;
	r = netlink_senddsbmsg(storage, &dsbmsg);
	fprintf(out, "(%d)DONE.\n", r);

	return;
}

static void __usercap_change(FILE *out, dsb_secstore_t *storage, uid_t uid, uint8_t n)
{
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
	dsbmsg.cmd = DSBNL_CMDUIDCAPCHANGE;
	dsbmsg.uid = uid;
	dsbmsg.listno = n;
	r = netlink_senddsbmsg(storage, &dsbmsg);
	fprintf(out, "(%d)DONE.\n", r);

	return;
}

static void __usercap_remove(FILE *out, dsb_secstore_t *storage, uid_t uid)
{
	struct dsbnl_policy_msg dsbmsg;
	int r = 0;

	memset(&dsbmsg, 0, sizeof(struct dsbnl_policy_msg));
	dsbmsg.cmd = DSBNL_CMDUIDCAPREMOVE;
	dsbmsg.uid = uid;
	r = netlink_senddsbmsg(storage, &dsbmsg);
	fprintf(out, "(%d)DONE.\n", r);

	return;
}

/*
 * ok, here the best place to describe S-expression used for control all
 * the stuff.
 * 1. Listing - show the list of already configured object permissions
 *    '(list-all)' - will reply with all objects
 *    '(list <hash>)' - will show full information about object permissions
 * 2. Adding objects to the list
 *    '(object-add (:blkid <128bit identifier>)(:inode <64bit long>)(:name "<name>"))'
 * 3. Removing objects from the list
 *    '(object-remove <hash>)'
 * 4. Edit rules of the object
 *    (object-entry-set-uid <hash> (:n 0)(:id 1000)(:mode rwx))
 *    (object-entry-set-gid <hash> (:n 0)(:id 1000)(:mode rwx))
 *    (object-entry-set-other <hash> (:mode rwx))
 *  or, in case of entry clear
 *    '(object-entry-clean-uid <hash> (:n 0))'
 *    '(object-entry-clean-gid <hash> (:n 0))'
 *    '(object-entry-clean-other <hash>)'
 */
static int __process_msg(FILE *out, char *msg, struct dsbd_ctx *gctx)
{
	/* ok, S-expressions again */
	sexp_t *sx = NULL, *isx, *iisx;
	dsb_secstore_t *storage = gctx->storage;
	uint64_t hash = 0;
	uint64_t inum;
	uint8_t blkid[16], mask;
	char *name, *fu_name, *val_name;
	uid_t uid;
	int idx, iidx, i, n, rid;
	char _des[5];

	if(!msg || !strlen(msg)) return 0;

	if(!(sx = parse_sexp(msg, strlen(msg)))) return 0;

	if(sx->ty == SEXP_LIST)
		fu_name = sx->list->val;
	else goto __fini;

	if(!fu_name || !strlen(fu_name)) goto __fini;

	/* now we're ready to call function */
	if(!strcmp(fu_name, "list-all")) { /* no parse at all */
		__print_active_object_from_storage(out, storage);
	} else if(!strcmp(fu_name, "list")) { /* need to take a hash */
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(isx->ty == SEXP_LIST) goto __fini;
			if(idx == 1) hash = strtoul(isx->val, NULL, 0);
		}

		__print_object_entries(out, storage, hash);
	} else if(!strcmp(fu_name, "object-add")) {
		name = NULL; inum = 0;
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx && isx->ty != SEXP_LIST) goto __fini; /* skip */

			if(idx) {
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */

					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}

				if(!strcmp(val_name, ":blkid")) {
					memset(blkid, 0, sizeof(uint8_t)*16);

					/* convert it to normal base */
					for(iidx = 0, i = 0; name[iidx] != '\0'; i++, iidx += 2) {
						if(i > 15) goto __fini;
						memset(_des, 0, 5);
						snprintf(_des, 5, "0x%c%c", name[iidx], name[iidx + 1]);
						blkid[i] = strtoul(_des, NULL, 0);
					}
				} else if(!strcmp(val_name, ":inode")) inum = strtoul(name, NULL, 10);
				else if(!strcmp(val_name, ":name")) name = name; /* nothing to do */
				else goto __fini;
			}
		}

		if(!name || !inum) goto __fini;

		__add_object_storage(out, storage, name, inum, blkid);
	} else if(!strcmp(fu_name, "object-remove")) {
		hash = 0;
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(isx->ty == SEXP_LIST) goto __fini;
			if(idx == 1) hash = strtoul(isx->val, NULL, 0);
		}

		if(!hash) goto __fini;
		__remove_object_storage(out, storage, hash);
	} else if(!strcmp(fu_name, "object-entry-set-uid") || !strcmp(fu_name, "object-entry-set-gid")) {
		/* rid */
		if(!strcmp(fu_name, "object-entry-set-uid")) rid = DSBRID_UID;
		else rid = DSBRID_GID;

		hash = 0; n = 0; mask = 0;
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx < 2 && isx->ty == SEXP_LIST) goto __fini;
			if(idx > 1 && isx->ty != SEXP_LIST) goto __fini;

			switch(idx) {
			case 0: break; /* check here ? */
			case 1: hash = strtoul(isx->val, NULL, 0); break;
			case 2:
			case 3:
			case 4:
			case 5:
				name = NULL;
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */

					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}

				if(!name) goto __fini;

				/* take care about what we got */
				if(!strcmp(val_name, ":n")) n = atoi(name);
				else if(!strcmp(val_name, ":mode")) {
					mask = 0;
					if(strlen(name) != 3) goto __fini;

					for(i = 0; i < 3; i++) if(name[i] != '-') mask |= (1 << (i + 1));
				} else if(!strcmp(val_name, ":id")) uid = atoi(name);
				else goto __fini;

				break;
			default: goto __fini;
			}
		}

		if(!hash) goto __fini;

		__setrule_object(out, storage, hash, n, uid, rid, mask);
	} else if(!strcmp(fu_name, "object-entry-set-other")) {
		rid = DSBRID_OTHER;
		hash = 0; mask = 0;
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx < 2 && isx->ty == SEXP_LIST) goto __fini;
			if(idx > 1 && isx->ty != SEXP_LIST) goto __fini;

			switch(idx) {
			case 0: break; /* check here ? */
			case 1: hash = strtoul(isx->val, NULL, 0); break;
			case 2:
			case 3:
				name = NULL;
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */

					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}

				if(!name) goto __fini;

				/* take care about what we got */
				if(!strcmp(val_name, ":mode")) {
					mask = 0;
					if(strlen(name) != 3) goto __fini;

					for(i = 0; i < 3; i++) if(name[i] != '-') mask |= (1 << (i + 1));
				} else goto __fini;

				break;
			default: goto __fini;
			}
		}

		if(!hash) goto __fini;

		__setrule_object(out, storage, hash, 0, 0, rid, mask);
	} else if(!strcmp(fu_name, "object-entry-clean-uid") || !strcmp(fu_name, "object-entry-clean-gid")) {
		/* rid */
		if(!strcmp(fu_name, "object-entry-clean-uid")) rid = DSBRID_UID;
		else rid = DSBRID_GID;
		hash = 0; n = 0;

		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx < 2 && isx->ty == SEXP_LIST) goto __fini;
			if(idx > 1 && isx->ty != SEXP_LIST) goto __fini;

			switch(idx) {
			case 0: break; /* check here ? */
			case 1: hash = strtoul(isx->val, NULL, 0); break;
			case 2:
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */

					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}

				/* take care about what we got */
				if(!strcmp(val_name, ":n")) n = atoi(name);
				else goto __fini;

				break;
			default: goto __fini;
			}
		}

		if(!hash) goto __fini;

		__cleanrule_object(out, storage, hash, n, rid);
	} else if(!strcmp(fu_name, "object-entry-clean-other")) {
		rid = DSBRID_OTHER;
		hash = 0;
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx < 2 && isx->ty == SEXP_LIST) goto __fini;
			if(idx > 1 && isx->ty != SEXP_LIST) goto __fini;

			switch(idx) {
			case 0: break; /* check here ? */
			case 1: hash = strtoul(isx->val, NULL, 0); break;
			default: goto __fini;
			}
		}

		if(!hash) goto __fini;

		__cleanrule_object(out, storage, hash, 0, rid);
	} else if(!strcmp(fu_name, "user-cap-set") ||
						!strcmp(fu_name, "user-cap-change")) { /* (user-cap-set|change (:uid <uid>) (:mask <mask>)) */
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx && isx->ty != SEXP_LIST) goto __fini;

			if(idx) {
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */
					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}
				if(!strcmp(val_name, ":uid")) uid = atoi(name);
				else if(!strcmp(val_name, ":mask")) n = atoi(name);
				else goto __fini;
			}

		}

		if(!strcmp(fu_name, "user-cap-set")) __usercap_set(out, storage, uid, n);
		else __usercap_change(out, storage, uid, n);
	} else if(!strcmp(fu_name, "user-cap-remove")) { /* (user-cap-set (:uid <uid>)) */
		SEXP_ITERATE_LIST(sx, isx, idx) {
			if(idx && isx->ty != SEXP_LIST) goto __fini;

			if(idx) {
				SEXP_ITERATE_LIST(isx, iisx, iidx) {
					if(iisx->ty == SEXP_LIST) goto __fini; /* skip */
					switch(iidx) {
					case 0: val_name = iisx->val; break;
					case 1: name = iisx->val; break;
					default: goto __fini; break;
					}
				}
				if(!strcmp(val_name, ":uid")) uid = atoi(name);
				else goto __fini;
			}

		}

		__usercap_remove(out, storage, uid);
	} else goto __fini;

 __fini:
	destroy_sexp(sx);

	return 0;
}

static void *__ctrl_thread(void *arg)
{
	struct dsbd_ctx *gctx = (struct dsbd_ctx *)arg;
	struct _payload *pload;
	zst_msg_t msg;
	FILE *wstream = NULL;

	while(1) {
		if(zst_queue_get(gctx->uqueue, NULL, &msg)) break;

		if(msg.msgtype != USR_MSG) continue;
		else pload = (struct _payload *)msg.data;

		wstream = fdopen(pload->scn, "r+");
		if(!wstream) {
			printf("errno = %s\n", strerror(errno));
			close(pload->scn);
			goto __fini_msg;
		}

		__process_msg(wstream, pload->pmsg, gctx);

		fflush(wstream);
		fclose(wstream);

	__fini_msg:
		/* finish up the message */
		//free(pload->pmsg);
		//free(pload);
		wstream = NULL;
	}

	return NULL;
}

/* peeeeeee
 * have a nice days until systemd isn't goes deep
 */
int main(int argc, char **argv)
{
	int c, r = 0, fp = 0;
	struct dsbd_ctx *gctx = malloc(sizeof(struct dsbd_ctx));
	dsb_secstore_t *storage = malloc(sizeof(dsb_secstore_t));
	zst_queue_t *queue = malloc(sizeof(zst_queue_t));
	pthread_t unixsck_thread, cmds_thread;

	if(!queue) return -ENOMEM;
	if(!gctx || !storage) return -ENOMEM;
	else { /* initial */
		memset(gctx, 0, sizeof(struct dsbd_ctx));

		if((r = dsb_store_allocinit(storage))) goto __fails;
		else gctx->storage = storage;

		if((r = zst_queue_init(queue))) goto __fails;
		else gctx->uqueue = queue;
	}

	/* this is time to parse parameters */
	while (1) {
    static struct option long_options[] = {
      /* These options a generic ones. */
      {"help", no_argument, NULL, 'h'}, /* print out help and version info */
      {"version", no_argument, NULL, 'v'}, /* just out a version info */
      /*  The following options about system. */
      {"daemonize", no_argument, NULL, 'd'},
			{"systemd", no_argument, NULL, 's'},
      {"unix-socket-path", required_argument, NULL, 'u'},
      {"log-file-path", required_argument, NULL, 'l'},
      {NULL, 0, NULL, 0},
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    if((c = getopt_long(argc, argv, "hvdsu:l:", long_options,
                        &option_index)) == -1) break;

    switch (c) {
    case 'h':
      __print_help();
      return 0;
      break;
    case 'v':
      __print_version();
      return 0;
      break;
    case 'd':
      gctx->flags |= DSBD_DAEMONIZE;
      break;
    case 's':
      gctx->flags |= DSBD_SYSTEMSHIT;
      break;
    case 'u':
			if(!(gctx->unixsck_path = strdup(optarg)))				{ r = -ENOMEM; goto __fails; }
			fp++;
      break;
    case 'l':
			if(!(gctx->log_file = strdup(optarg)))				{ r = -ENOMEM; goto __fails; }
			fp++;
      break;
    case '?':
      /* getopt_long already printed an error message. */
      break;
    default:
      abort();
    }
  }

	if(fp < 2) {
		fprintf(stderr, "FAIL: No file pathes provided, not enough options.\n");

		__print_version();
		__print_help();
		return -EINVAL;
	}

	if(gctx->flags & DSBD_DAEMONIZE) skeleton_daemon();

	/* ok now we're ready to do the other initialization */
	dsb_sectore_setinitdac(gctx->storage, store_initdac);
	dsb_sectore_setinitbe(gctx->storage, store_initbe);
  dsb_sectore_setpushdac(gctx->storage, store_pushbe);
  dsb_sectore_setpushbe(gctx->storage, store_pushdac);

	r = dsb_store_init(gctx->storage, gctx, gctx);
	if(r) goto __fails;

	/* firstly open unix socket to listen for cmds */
	/*
	 * NOTE: this part might be completely changed for more
	 * appropriate way to make it so.
	 * my suggestion to use special secure local bus, no no no
	 * no dbus, just something lightweight and secure. Proposal required.
	 */
	if(pthread_create(&unixsck_thread, NULL, __unixsocket_thread, (void *)gctx)) {
		syslog(LOG_CRIT, "Failed to create a special unix socket listening thread.\n");
		return -ENOMEM;
	}
	if(pthread_create(&cmds_thread, NULL, __ctrl_thread, (void *)gctx)) {
		syslog(LOG_CRIT, "Failed to create a special unix socket command thread.\n");
		return -ENOMEM;
	}

	/* main thread will works with netlink */
	int infd = scknl_openevents();
	if(infd < 0) {
		for(;;) {
			sleep(10000);
		}
	} else {
		events_loop(stderr, gctx, infd);
	}

	pthread_cancel(unixsck_thread);
	pthread_cancel(cmds_thread);

  return 0;

 __fails:
	free(gctx);
	free(storage);

	return -r;
}
