/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * Originally written by Alexander Vdolainen <avdolainen@gmail.com>
 */

#include <errno.h>
#include <string.h>
#include <dsbus/data.h>

static long __cmpuint64(const void *a, const void *b);

int dsb_store_allocinit(dsb_secstore_t *storage)
{
  if(!storage) return EINVAL;
  else memset(storage, 0, sizeof(dsb_secstore_t));

  if(pthread_rwlock_init(&(storage->srwlock), NULL)) return ENOMEM;

  usrtc_init(&(storage->curr_data), USRTC_SPLAY, MAX_CTRLED_OBJECTS,
             __cmpuint64);
  usrtc_init(&(storage->garbaged_data), USRTC_SPLAY, MAX_CTRLED_OBJECTS,
             __cmpuint64);

  return 0;
}

int dsb_store_init(dsb_secstore_t *storage, void *beptr, void *dacptr)
{
  int r = 0;
  if(!storage->store_initbe || !storage->store_initdac) return EINVAL;

  if((r = storage->store_initbe(storage, beptr))) return r;
  if((r = storage->store_initdac(storage, dacptr))) return r;

  return r;
}

int dsb_store_flush(dsb_secstore_t *storage)
{
  usrtc_t *stree = &(storage->curr_data);
  usrtc_node_t *node = NULL;
  dsb_secinfo_t *info = NULL;
  int r = 0;

  /* globally lock it */
  pthread_rwlock_wrlock(&(storage->srwlock));

  for(node = usrtc_first(stree); node != usrtc_last(stree);
      node = usrtc_next(stree, node)) {
    info = (dsb_secinfo_t *)usrtc_node_getdata(node);

    pthread_mutex_lock(&info->slock);
    if((info->flags & DSB_DIRTY) && (info->flags & DSB_CONFIGURED)) {
      r = dsb_sectore_pushbe(storage, info);
      if(!r) r = dsb_sectore_pushdac(storage, info);

			/* clear flag */
      if(!r) info->flags &= ~DSB_DIRTY;
    }
    pthread_mutex_unlock(&info->slock);

    if(r) break;
  }

  pthread_rwlock_unlock(&(storage->srwlock));

  return r;
}

int dsb_store_clean(dsb_secstore_t *storage)
{
  usrtc_t *stree = &(storage->garbaged_data);
  dsb_secinfo_t *info = NULL;
  usrtc_node_t *node = NULL;
  int r = 0;

  /* globally lock it */
  pthread_rwlock_wrlock(&(storage->srwlock));
  for(node = usrtc_first(stree); node != NULL;
      node = usrtc_first(stree)) {
    info = (dsb_secinfo_t *)usrtc_node_getdata(node);

    usrtc_delete(stree, node);
    if((r = dsb_store_info_destroy(info))) break;
  }

  if(!r) {
    stree = &(storage->curr_data);

    for(node = usrtc_first(stree); node != usrtc_last(stree);
        node = usrtc_next(stree, node)) {
      info = (dsb_secinfo_t *)usrtc_node_getdata(node);

      if((info->flags & DSB_GARBAGED) && (info->flags & DSB_CONFIGURED)) {
        pthread_mutex_lock(&info->slock);
        if((r = dsb_sectore_pushbe(storage, info))) goto __l_fail;
        if((r = dsb_sectore_pushdac(storage, info))) goto __l_fail;

        /* move node to the garbage */
        usrtc_delete(stree, node);
        usrtc_insert(&(storage->garbaged_data), node, &(info->hash));

      __l_fail:
        pthread_mutex_unlock(&info->slock);
        if(r) break;
      }
    }
  }

  pthread_rwlock_unlock(&(storage->srwlock));

  return r;
}

int dsb_store_insert(dsb_secstore_t *storage, dsb_secinfo_t *info)
{
	usrtc_t *stree = &(storage->curr_data);
  usrtc_node_t *node = NULL;
	int r = 0;

	if(!info) return EINVAL;

	/* check for duplicates */
	pthread_rwlock_rdlock(&(storage->srwlock));
	if(usrtc_count(stree) >= MAX_CTRLED_OBJECTS) r = EXFULL;

	node = usrtc_lookup(stree, &(info->hash));
	if(node) r = EEXIST;
	pthread_rwlock_unlock(&(storage->srwlock));

	if(r) return r;
	else node = &(info->node);

	/* global lock */
	pthread_rwlock_wrlock(&(storage->srwlock));
	usrtc_insert(stree, node, &(info->hash));
	pthread_rwlock_unlock(&(storage->srwlock));

  return 0;
}

int dsb_store_remove(dsb_secstore_t *storage, uint64_t hash)
{
	usrtc_t *stree = &(storage->curr_data);
  usrtc_node_t *node = NULL;
	dsb_secinfo_t *info;
	int r = 0;

	pthread_rwlock_rdlock(&(storage->srwlock));
	if(usrtc_count(stree) >= MAX_CTRLED_OBJECTS) r = EXFULL;

	node = usrtc_lookup(stree, &hash);
	if(!node) r = ENOENT;
	pthread_rwlock_unlock(&(storage->srwlock));

	if(!r) {
		info = (dsb_secinfo_t *)usrtc_node_getdata(node);
		pthread_mutex_lock(&(info->slock));
		info->flags |= DSB_GARBAGED;
		info->flags |= DSB_DIRTY;
		pthread_mutex_unlock(&(info->slock));
	}

  return r;
}

int dsb_store_search(dsb_secstore_t *storage, uint64_t hash,
                     dsb_secinfo_t **info)
{
	usrtc_t *stree = &(storage->curr_data);
  usrtc_node_t *node = NULL;
	int r = 0;

	pthread_rwlock_rdlock(&(storage->srwlock));
	node = usrtc_lookup(stree, &hash);
	if(!node) r = ENOENT;

	pthread_rwlock_unlock(&(storage->srwlock));

	if(!r)		*info = (dsb_secinfo_t *)usrtc_node_getdata(node);

  return r;
}

/* --- infos --- */

int dsb_store_info_allocinit(dsb_secinfo_t *info)
{
	if(!info) return EINVAL;

	memset(info, 0, sizeof(dsb_secinfo_t));

	if(pthread_mutex_init(&(info->slock), NULL)) return ENOMEM;

	usrtc_node_init(&(info->node), (void *)info);

  return 0;
}

int dsb_store_info_destroy(dsb_secinfo_t *info)
{
	if(!info) return EINVAL;

	pthread_mutex_destroy(&(info->slock));

  return 0;
}

struct storage_backend {
	int fd;
};

int store_initbe(dsb_secstore_t *storage, void *priv)
{
	struct storage_backend *be = malloc(sizeof(struct storage_backend));

	if(!be) return ENOMEM;
	else storage->store_be = be;

	return 0;
}

int store_pushbe(dsb_secstore_t *storage, dsb_secinfo_t *info)
{
	/* dummy */
	return 0;
}

static long __cmpuint64(const void *a, const void *b)
{
  return *(uint64_t *)a - *(uint64_t *)b;
}
